% complementary nucleotides
cmpltry(a,u).
cmpltry(u,a).
cmpltry(c,g).
cmpltry(g,c).

% complementary sequences
cmpltry([H1|T1], [H2|T2]) :- cmpltry(H1,H2), cmpltry(T1,T2).
cmpltry([H1], [H2]) :- cmpltry(H1,H2).

% non-complementary nucleotides
ncmpltry(a,a).
ncmpltry(a,c).
ncmpltry(a,g).
ncmpltry(u,u).
ncmpltry(u,c).
ncmpltry(u,g).
ncmpltry(g,a).
ncmpltry(g,g).
ncmpltry(g,u).
ncmpltry(c,c).
ncmpltry(c,a).
ncmpltry(c,u).

% writer(A) :- ncmpltry(A, X), write(X),! , B != A, writer(B).

% non-complementary sequences
ncmpltry([H1|T1], [H2|T2]) :- ncmpltry(H1,H2), ncmpltry(T1,T2).
ncmpltry([H1], [H2]) :- ncmpltry(H1,H2).

% Yes if heads are equal
equalheads([X|T1], [X|T2]) :- equalheads(T1, T2).
equalheads([], _).

% Yes if B contains A
contains([X|T1], [X|T2]) :- equalheads(T1, T2).
contains([X|T1], [_|T2]) :- contains([X|T1], T2).

% detects basic bubbles like this:
% buble([a,a,a,c,g,a,t,a,g,c,a,t], X).
%
buble([A1,A2,A3,C1,C2,C3,A3,A2,A1|T], [[C1,C2,C3]|T]).
buble([H|T],[H|X]) :- buble(T,X).

% Yes if heads are equal, C contains nonequal tail
equalheads([X|T1], [X|T2], Y) :- atom(X), equalheads(T1, T2, Y).
equalheads([], X, X).

% Yes if B contains A, C has B with A cutted off
contains([X|T1], [X|T2], [], T3) :- equalheads(T1, T2, T3).
contains([X|T1], [Y|T2], [Y|H3], T3) :- contains([X|T1], T2, H3, T3).

% pushes Head list items to Tail list in Result
% push(Head, Tail, Result)
push([H|T1], T2, [H|T3]) :- push(T1, T2, T3).
push([], T, T).

% joins A B C D E into one Result list with order [A,B,C,D,E]
join(A, B, C, D, E, Result) :- push(E, [], R1), push(D, R1, R2), push(C, R2, R3), push(B, R3, R4), push(A, R4, Result).

% substring A in S, result R
% first argument is at least 2 long -> complimentary sequences are at least 2 long
% ex: bubble([_,_|_], [a,g,g,g,c,a,u,g,c,c,c,g,a],Result).
bubble(A,S,Result) :- 
	contains(A, S, Pref1, Postf1), 
	cmpltry(A, ComplA), reverse(ComplA,ComplRevA), 
	contains(ComplRevA, Postf1, Pref2, Postf2), 
	join(Pref1, [A], Pref2, [ComplRevA], Postf2, Result).

% pure idea
idealbubble(S) :- 
	contains(A, S, _, Tail), 
	contains(A, Tail, _, _).

% true if argument 2 is a list of at least K elements
atleast(K,[_|T]) :- K1 is K-1,atleast(K1,T).
atleast(1,[_|_]).

% all structures
fold(S,Result,N) :-
	contains(A, S, Pref1, Postf1),
	atleast(4, A),
	cmpltry(A, ComplA), reverse(ComplA,ComplRevA), 
	contains(ComplRevA, Postf1, Pref2, Postf2),
	atleast(4, Pref2),
	join(Pref1, [[N,A]], Pref2, [[N,ComplRevA]], Postf2, R),
	NewN is N+1,
	fold(R,Result,NewN).
fold(X, X, _).

% tree structure
fold3(S,Result) :-
	contains(A, S, Pref1, Postf1),
	atleast(4, A),
	cmpltry(A, ComplA), reverse(ComplA,ComplRevA), 
	contains(ComplRevA, Postf1, Pref2, Postf2),
	atleast(4, Pref2),
	fold3(Pref1, Pref1F),
	fold3(Pref2, Pref2F),
	fold3(Postf2, Postf2F),
	N is random(10000),
	join(Pref1F, [[N,A]], Pref2F, [[N,ComplRevA]], Postf2F, Result).

fold3(X, X).