#include "def.h"

int init()
{
  pvm_joingroup(GROUPNAME);
  pvm_barrier(GROUPNAME, N);
  return pvm_getinst(GROUPNAME, pvm_mytid());
}

void leave()
{
  printf("leaving\n"); fflush(stdout);
  //  pvm_barrier(GROUPNAME, N);
  pvm_lvgroup(GROUPNAME);
  pvm_exit();
}

void send(int from, int to, int type, int value)
{
  printf("%d send to %d: %s(%d)\n", from, to, msg_type[type], value); fflush(stdout);

  int to_tid = pvm_gettid(GROUPNAME, to);
  if (to_tid < 0)
    leave();

  pvm_initsend(PvmDataDefault);
  pvm_pkint(&from, 1, 1);
  pvm_pkint(&type, 1, 1);
  pvm_pkint(&value, 1, 1);
  pvm_send(to_tid, MSG_SLV);
}

void recv(int *from, int *type, int *value)
{
  pvm_recv( -1, MSG_SLV);
  pvm_upkint(from, 1, 1);
  pvm_upkint(type, 1, 1);
  pvm_upkint(value, 1, 1);
}

int remove_max_index(int* array)
{
  int i;
  int index = -1;
  for (i=0; i< N; i++) {
    if (array[i] >= 0 && ((index == -1) || array[i] >= array[index]))
      index = i;
  }
  array[index] = -1;
  return index;
}

void reset(int* array)
{
  int i;
  for (i=0; i<N; i++) {
    array[i] = -1;
  }
}

main()
{
  int i = init();

  time_t deadline = time(NULL) + 60;

  int state = initial_state[i];

  int blocked[N];
  int bcount = 0;
  int dest = i;

  int recv_type;
  int recv_val;
  int recv_from;

  int token_val = 0;
  int b;

  reset(blocked);

  while (time(NULL) < deadline) {
    switch (state) {

    case (EMPTY):
      recv(&recv_from, &recv_type, &recv_val);
      send(i, recv_from, ACK, i);
      state = WAITING_TOKEN;
      break;

    case (WAITING_TOKEN):
      recv(&recv_from, &recv_type, &recv_val);
      switch (recv_type) {
      case (REQ):
	send(i, recv_from, BUSY, -1);
	break;
      case (TOKEN):
	token_val = recv_val + 1;
	dest = (i + 1) % N;
	bcount = 0;
	reset(blocked);
	send(i, dest, REQ, token_val);
	state = WAITING_ANSWER;
	break;
      }
      break; // case (WAITING_TOKEN)

    case (WAITING_ANSWER):
      recv(&recv_from, &recv_type, &recv_val);
      switch (recv_type) {
      case (BUSY):
	do { // zwiekszaj, pomijajac blokowanych
	  dest = (dest + 1) % N;
	} while (blocked[dest] >= 0);
	if (dest != i) {
	  send(i, dest, REQ, token_val);
	} else if (bcount > 0) { // sprobuj jeszcze raz, jesli sa jacys zablokowani i juz skonczyles
	    printf("%d NEW TRY, bcount=%d, i will ask ", i, bcount);
	    dest = (i +1) % N;
	    while (blocked[dest] >= 0) { // zwiekszaj, pomijajac blokowanych
		dest = (dest + 1) % N;
	    }
	    printf("%d\n", dest); fflush(stdout);
	    send(i, dest, REQ, token_val);
	}
	break;

      case (ACK):
	send(i, recv_val, TOKEN, token_val);

	while (bcount > 1) {
	  b = remove_max_index(blocked);
	  if (b != -1)
	    send(i, b, BUSY, -1);
	  bcount--;
	}
	bcount = 0;

	b = remove_max_index(blocked);
	if (b != -1) {
	  printf("%d awakens BLOCKED %d by ACK\n", i, b); fflush(stdout);
	  send(i, b, ACK, i);
	  state = WAITING_TOKEN;
	} else { // bcount == 0
	  state = EMPTY;
	}
	break;

      case (REQ):
	if (recv_val <= token_val) {
	  send(i, recv_from, BUSY, token_val);
	} else {
	  bcount++;
	  blocked[recv_from] = recv_val;
	  printf("%d BLOCK %d with value %d (my is %d), total %d blocked (ja wyslalem do %d)\n", i, recv_from, recv_val, token_val, bcount, dest); fflush(stdout);
	 
	  if (dest == recv_from || dest == i) { // od tego co mu wyslalismy
	    dest = (dest + 1) % N;
	    while (blocked[dest] >= 0 || dest == i) {
	     // zwiekszaj, pomijajac blokowanych
	      dest = (dest + 1) % N;
	    }
	    printf("%d REQ from DEST(%d), restart from %d\n", i, recv_from, dest); fflush(stdout);
	    send(i, dest, REQ, token_val);
	  }
	 
	}
	break;
      }
      break; // case (WAITING_ANSWER)

    case (HAS_TOKEN):
      dest = (dest + 1) % N;
      send(i, dest, REQ, token_val);
      state = WAITING_ANSWER;
      break;
    }
    }

  printf("TIMEOUT TIMEOUT TIMEOUT\n"); fflush(stdout);
  leave();
}

