#include "def.h"

int connection_init()
{
  srand(pvm_mytid());
  pvm_joingroup(GROUPNAME);
  pvm_barrier(GROUPNAME, N);
  return pvm_getinst(GROUPNAME, pvm_mytid());
}

void connection_leave()
{
  printf("leaving\n"); fflush(stdout);
  //  pvm_barrier(GROUPNAME, N);
  pvm_lvgroup(GROUPNAME);
  pvm_exit();
}

void send(int from, int to, int type, int value)
{
  //printf("%d send to %d: %s(%d)\n", from, to, msg_type[type], value); fflush(stdout);

  int to_tid = pvm_gettid(GROUPNAME, to);
  if (to_tid < 0)
    connection_leave();

  pvm_initsend(PvmDataDefault);
  pvm_pkint(&from, 1, 1);
  pvm_pkint(&type, 1, 1);
  pvm_pkint(&value, 1, 1);
  pvm_send(to_tid, MSG_SLV);
}

void recv(recv_t* msg)
{
  pvm_recv( -1, MSG_SLV);
  pvm_upkint(&msg->from, 1, 1);
  pvm_upkint(&msg->type, 1, 1);
  pvm_upkint(&msg->value, 1, 1);
}

int remove_min_index(int* array)
{
  int i;
  int index = -1;
  for (i=0; i< N; i++) {
    if (array[i] > -1 && ((index == -1) || array[i] <= array[index]))
      index = i;
  }

  if (index != -1) {
    array[index] = -1;
  }

  return index;
}

/* // RANDOM_SELECTION
int is_next_available(int i, int* blocked, int* visited)
{
  int c = i;
  do {
    c = (c + 1) % N;
  } while ((i != c) && ((blocked[c] > -1) || (visited[c] > -1)));

  return (i != c);
}

int next_available(int i, int* blocked, int* visited)
{
  int c = rand() % N;

  while ((c == i) || (blocked[c] > -1) || (visited[c] > -1)) {
    c = (c + 1) % N;
  }
  
  visited[c] = 1;
  return c;
}
*/


int is_next_available(int i, int current, int* blocked)
{
  do {
    current = (current + 1) % N;
  } while ((i!= current) && (blocked[current] > -1));

  return (i != current);
}

int next_available(int current, int* blocked)
{
  do {
    current = (current + 1) % N;
  } while (blocked[current] > -1);

  return current;
}

void reset(int* array)
{
  int i;
  for (i=0; i<N; i++) {
    array[i] = -1;
  }
}

//void init(int i, int* blocked, int* visited, int* dest, int* retry)
void init(int i, int* blocked, int* dest, int* retry)
{
  //printf("%d init\n", i); fflush(stdout);
  reset(blocked);
  // reset(visited); // CHANGE
  *retry = 0;
  *dest = next_available(i, blocked);
  //*dest = next_available(i, blocked, visited); // CHANGE
}

int blocks(int* blocked)
{
  int i;
  for (i=0; i<N; i++) {
    if (blocked[i]> -1) 
      return 1;
  }

  return 0;
}

void release_blocked(int i, int* blocked)
{
  int b = remove_min_index(blocked);
  printf("%d release blocked\n", i); fflush(stdout);
  //printf("%d release blocked: %d ", i, b);

  if (b != -1) {
    send(i, b, ACK, i);
  }

  while ((b = remove_min_index(blocked)) != -1) {
    //printf("%d ", b);
    send(i, b, BUSY, i);
  }
  //printf("\n"); fflush(stdout);
}

main()
{
  int i = connection_init();

  time_t timeout = time(NULL) + 100;

  int state = initial_state[i];

  int blocked[N];
  //int visited[N]; // CHANGE
  int dest = i;

  recv_t received;

  int token_val = 0;
  int retry = 0;


  while (time(NULL) <= timeout) { // while not end of game
    //printf("%d state %s\n", i, state_name[state]); fflush(stdout);

    switch (state) {
    case (HAS_TOKEN) :
    case (INIT) :
      printf("%d has token (%d)\n", i, token_val); fflush(stdout);
      //      init(i, blocked, visited, &dest, &retry); // CHANGE
      init(i, blocked, &dest, &retry);
      state = SEND_REQ;
      break;

    case (SEND_REQ) :
      send(i, dest, REQ, token_val);
      state = RECV_ANSWER;
      break;

    case (RECV_ANSWER) :
      recv(&received);
      switch (received.type) {
      case (ACK) :
	send(i, dest, TOKEN, token_val);
	//printf("%d check if blocks? %d\n", i, blocks(blocked)); fflush(stdout);;
	if (blocks(blocked)) {
	  release_blocked(i, blocked);
	  state = RECV_TOKEN;
	} else {
	  state = RECV_REQ;
	}
	break;
	
      case (REQ) :
	if (received.value > token_val) {
	  blocked[received.from] = received.value;
	  retry++;
	  send(i, dest, NOTIFY, -1);
	  printf("%d block %d with %d, my is %d (retr=%d)\n", i, received.from, received.value, token_val, retry); fflush(stdout);
	  state = RECV_ANSWER;
	} else {
	  send(i, received.from, BUSY, -1);
	  state = RECV_ANSWER;
	}
	break;

      case (NOTIFY) :
	retry++;
	state = RECV_ANSWER;
	break;

      case (BUSY) :
	//if (is_next_available(i, blocked, visited)) {
	//dest = next_available(i, blocked, visited);
	if (is_next_available(i, dest, blocked)) { // CHANGE
	  dest = next_available(dest, blocked); // CHANGE
	  state = SEND_REQ;
	} else {
	  if (retry == 0) {
	    printf("%d wait wakeup\n", i); fflush(stdout);
	    recv(&received); // wait wakeup
	    switch (received.type) {
	    case (NOTIFY) :
	      // empty case
	      break;

	    case (REQ) :
	      if (received.value > token_val) {
		blocked[received.from] = received.value;
		printf("%d block %d with %d, my is %d (go again)\n", i, received.from, received.value, token_val); fflush(stdout);
	      } else {
		send(i, received.from, BUSY, -1);
	      }
	      break;
	    }

	  } else {
	    printf("%d retry (%d retries total)\n",i, retry); fflush(stdout);
	    retry--;
	  }

	  // reset(visited); // CHANGE
	  // dest = next_available(i, blocked, visited);
	  dest = next_available(i, blocked); // CHANGE
	  state = SEND_REQ;
	}
	break;
      }
      break;

    case (EMPTY) :
    case (RECV_REQ) :
      printf("%d is empty\n", i); fflush(stdout);
      recv(&received);
      switch (received.type) {
      case (NOTIFY) :
	// no code
	break;

      case (REQ) :
	send(i, received.from, ACK, i);
	state = RECV_TOKEN;
	break;

      }
      break;

    case (RECV_TOKEN) :
      recv(&received);
      switch (received.type) {
      case (NOTIFY) :
	// no code
	break;

      case (REQ) :
	send(i, received.from, BUSY, -1);
	break;

      case (TOKEN) :
	token_val = received.value + 1;
	state = INIT;
	break;

      }
      break;
      
    }
  }

  printf("TIMEOUT TIMEOUT TIMEOUT\n"); fflush(stdout);
  connection_leave();
}
