#include <stdio.h>
#include <string.h>

#define ACK 0
#define BUSY 1

#define N 10

void send(int from, int to, int type, int value)
{
  printf("%d send to %d: %d(%d)\n", from, to, type, value); fflush(stdout);
}

int remove_min_index(int* array)
{
  int i;
  int index = -1;
  for (i=0; i< N; i++) {
    if (array[i] > -1 && ((index == -1) || array[i] <= array[index]))
      index = i;
  }

  if (index != -1) {
    array[index] = -1;
  }

  return index;
}

int is_next_available(int i, int current, int* blocked)
{
  do {
    current = (current + 1) % N;
  } while ((i!= current) && (blocked[current] > -1));

  return (i != current);
}

int next_available(int current, int* blocked)
{
  do {
    current = (current + 1) % N;
  } while (blocked[current] > -1);

  return current;
}

void reset(int* array)
{
  int i;
  for (i=0; i<N; i++) {
    array[i] = -1;
  }
}

void init(int i, int* blocked, int* dest, int* retry)
{
  printf("%d init\n", i); fflush(stdout);
  reset(blocked);
  *retry = 0;
  *dest = next_available(i, blocked);
}

int blocks(int* blocked)
{
  int i;
  for (i=0; i<N; i++) {
    if (blocked[i]> -1) 
      return 1;
  }

  return 0;
}

void release_blocked(int i, int* blocked)
{
  int b = remove_min_index(blocked);
  printf("%d releaseBlocked, min blocked = %d\n", i, b); fflush(stdout);

  if (b != -1) {
    send(i, b, ACK, i);
  }

  while ((b = remove_min_index(blocked)) != -1) {
    send(i, b, BUSY, i);
  }
}

void pra(int* array)
{
  int i;
  for (i=0; i<N; i++)
    printf("%d ", array[i]);
  printf("\n");
}

int main()
{
  int example[N] = {-1, -1, 3, 5, -1, 3, 5, -1, 8, 0};
  int array[N];
  int k, i, current, dest;

  memcpy(array, example, N*sizeof(int));
  printf("example: ");
  pra(example);
  printf("array: ");
  pra(array);

  printf("\n");

  printf("iterating available: \n");
  for (k = 0; k<N;  k++) {
    current = k;
    printf("\tcurrent=%d: ", current);
    i = 0;
    while (is_next_available(i, current, array)) {
      current = next_available(current, array);
      printf("%d ", current);
    }
    printf("\n");
  }

  printf("array: ");
  pra(array);

  printf("\n");

  printf("removing blocking: ");
  i = 0;
  current = 5;
  while (blocks(array)) {
    dest = remove_min_index(array);
    printf("%d ", dest);
  }
  printf("\narray: ");
  pra(array);
  
  printf("\n");
  
  memcpy(array, example, N*sizeof(int));
  printf("array: ");
  pra(array);
  printf("reset\n");
  reset(array);
  pra(array);

  memcpy(array, example, N*sizeof(int));
  printf("array: ");
  pra(array);
  printf("release blocked\n");
  release_blocked(0, array);
  pra(array);
  

  return 0;
}
